
/***************************************************************************
*
*   Copyright (c) 1998, 1999 Timpanogas Research Group, Inc.
*   895 West Center Street
*   Orem, Utah  84057
*   jmerkey@timpanogas.com
*
*   This program is free software; you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the
*   Free Software Foundation, version 2, or any later version.
*
*   This program is distributed in the hope that it will be useful, but
*   WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   General Public License for more details.
*
*   You are free to modify and re-distribute this program in accordance
*   with the terms specified in the GNU Public License.  The copyright
*   contained in this code is required to be present in any derivative
*   works and you are required to provide the source code for this
*   program as part of any commercial or non-commercial distribution.
*   You are required to respect the rights of the Copyright holders
*   named within this code.
*
*   jmerkey@timpanogas.com and TRG, Inc. are the official maintainers of
*   this code.  You are encouraged to report any bugs, problems, fixes,
*   suggestions, and comments about this software to jmerkey@timpanogas.com
*   or linux-kernel@vger.rutgers.edu.  New releases, patches, bug fixes, and
*   technical documentation can be found at www.timpanogas.com.  TRG will
*   periodically post new releases of this software to www.timpanogas.com
*   that contain bug fixes and enhanced capabilities.
*
*   Original Authorship      :
*      source code written by Jeff V. Merkey, TRG, Inc.
*
*   Original Contributors    :
*      Jeff V. Merkey, TRG, Inc.
*      Darren Major, TRG, Inc.
*      Alan Cox, RedHat Software, Inc.
*
****************************************************************************
*
*
*   AUTHOR   :  Jeff V. Merkey (jmerkey@timpanogas.com)
*   FILE     :  MMAP.C
*   DESCRIP  :  FENRIS VFS Memory Mapping Module for Linux
*   DATE     :  December 6, 1998
*
*
***************************************************************************/

#include "globals.h"

#if (LINUX_22)

inline int min(int a, int b)
{
    return a < b ? a : b;
}

ULONG nwfs_file_mmap_nopage(struct vm_area_struct *area,
			     unsigned long address, int no_share)
{
    struct inode * inode = area->vm_file->f_dentry->d_inode;
    unsigned long page;
    unsigned int clear;
    int pos;
    long gap;	// distance from eof to pos

    page = __get_free_page(GFP_KERNEL);
    if (!page)
		return page;
    address &= PAGE_MASK;
    pos = address - area->vm_start + area->vm_offset;

    clear = 0;
    gap = inode->i_size - pos;
    if (gap <= 0)
    {
       // mmaping beyond end of file
       clear = PAGE_SIZE;
    }
    else
    {
       int cur_read;
       int need_read;
       struct file filp;

       if (gap < PAGE_SIZE)
       {
	  clear = PAGE_SIZE - gap;
       }
       filp.f_reada = 0;
       filp.f_pos = pos;
       filp.f_dentry=area->vm_file->f_dentry;

       need_read = PAGE_SIZE - clear;
       {
	  mm_segment_t cur_fs = get_fs();
	  set_fs (KERNEL_DS);
	  cur_read = nwfs_file_read_kernel(&filp, (char *)page, need_read,
					   &filp.f_pos);
	  set_fs (cur_fs);
       }

       if (cur_read != need_read)
       {
	  NWFSPrint("nwfs: Error while reading an mmap file %d <> %d\n",
		    cur_read, need_read);
       }
    }

    if (clear > 0)
    {
       memset((char*)page+PAGE_SIZE-clear, 0, clear);
    }
    return page;
}

struct vm_operations_struct nwfs_file_mmap =
{
	NULL,			/* open */
	NULL,			/* close */
	NULL,			/* unmap */
	NULL,			/* protect */
	NULL,			/* sync */
	NULL,			/* advise */
	nwfs_file_mmap_nopage,/* nopage */
	NULL,			/* wppage */
	NULL,			/* swapout */
	NULL,			/* swapin */
};

int nwfs_mmap(struct file * file, struct vm_area_struct * vma)
{
    struct inode *inode = file->f_dentry->d_inode;

    if (vma->vm_flags & VM_SHARED)   // only PAGE_COW or read-only supported now
       return -EINVAL;
    if (vma->vm_offset & (inode->i_sb->s_blocksize - 1))
       return -EINVAL;
    if (!inode->i_sb || !S_ISREG(inode->i_mode))
       return -EACCES;
    if (!IS_RDONLY(inode))
    {
       inode->i_atime = NWFSGetSystemTime();
       mark_inode_dirty(inode);
    }

    vma->vm_ops = &nwfs_file_mmap;
    return 0;
}

#endif

#if (LINUX_20)

inline int min(int a, int b)
{
    return a < b ? a : b;
}

ULONG nwfs_file_mmap_nopage(struct vm_area_struct * area,
			     unsigned long address, int no_share)
{
    struct inode *inode = area->vm_inode;
    ULONG page;
    unsigned int clear;
    int pos;
    long gap;	/* distance from eof to pos */

    page = __get_free_page(GFP_KERNEL);
    if (!page)
		return page;
    address &= PAGE_MASK;
    pos = address - area->vm_start + area->vm_offset;

    clear = 0;
    gap = inode->i_size - pos;
    if (gap <= 0)
    {
       /* mmaping beyond end of file */
       clear = PAGE_SIZE;
    }
    else
    {
       int cur_read;
       int need_read;
       struct file filp;

       if (gap < PAGE_SIZE)
       {
	  clear = PAGE_SIZE - gap;
       }

       filp.f_reada = 0;
       filp.f_pos = pos;
       need_read = PAGE_SIZE - clear;
       {
	  ULONG cur_fs = get_fs();

	  set_fs (KERNEL_DS);
	  cur_read = nwfs_file_read(inode, &filp, (char *)page, need_read);
	  set_fs (cur_fs);
       }
       if (cur_read != need_read)
       {
	  NWFSPrint("nwfs: error reading mmap file %d <> %d\n", cur_read, need_read);
       }
    }
    if (clear > 0)
    {
       NWFSSet((char *)page + PAGE_SIZE - clear, 0, clear);
    }
    return page;
}

struct vm_operations_struct nwfs_file_mmap =
{
	NULL,			/* open */
	NULL,			/* close */
	NULL,			/* unmap */
	NULL,			/* protect */
	NULL,			/* sync */
	NULL,			/* advise */
	nwfs_file_mmap_nopage,/* nopage */
	NULL,			/* wppage */
	NULL,			/* swapout */
	NULL,			/* swapin */
};

int nwfs_mmap(struct inode * inode, struct file * file, struct vm_area_struct * vma)
{
    /* only PAGE_COW or read-only supported now */

    if (vma->vm_flags & VM_SHARED)
       return -EINVAL;

    if (!inode->i_sb || !S_ISREG(inode->i_mode))
       return -EACCES;

    if (!IS_RDONLY(inode))
    {
       inode->i_atime = NWFSGetSystemTime();
       inode->i_dirt = 1;
    }
    vma->vm_inode = inode;
    inode->i_count++;
    vma->vm_ops = &nwfs_file_mmap;
    return 0;
}

#endif
