
/***************************************************************************
*
*   Copyright (c) 1998, 1999 Timpanogas Research Group, Inc.
*   895 West Center Street
*   Orem, Utah  84057
*   jmerkey@timpanogas.com
*
*   This program is free software; you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the
*   Free Software Foundation, version 2, or any later version.
*
*   This program is distributed in the hope that it will be useful, but
*   WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   General Public License for more details.
*
*   You are free to modify and re-distribute this program in accordance
*   with the terms specified in the GNU Public License.  The copyright
*   contained in this code is required to be present in any derivative
*   works and you are required to provide the source code for this
*   program as part of any commercial or non-commercial distribution.
*   You are required to respect the rights of the Copyright holders
*   named within this code.
*
*   jmerkey@timpanogas.com and TRG, Inc. are the official maintainers of
*   this code.  You are encouraged to report any bugs, problems, fixes,
*   suggestions, and comments about this software to jmerkey@timpanogas.com
*   or linux-kernel@vger.rutgers.edu.  New releases, patches, bug fixes, and
*   technical documentation can be found at www.timpanogas.com.  TRG will
*   periodically post new releases of this software to www.timpanogas.com
*   that contain bug fixes and enhanced capabilities.
*
*   Original Authorship      :
*      source code written by Jeff V. Merkey, TRG, Inc.
*
*   Original Contributors    :
*      Jeff V. Merkey, TRG, Inc.
*      Darren Major, TRG, Inc.
*      Alan Cox, RedHat Software, Inc.
*
****************************************************************************
*
*
*   AUTHOR   :  Jeff V. Merkey (jmerkey@timpanogas.com)
*   FILE     :  NWVFS.C
*   DESCRIP  :  FENRIS VFS Module for Linux
*   DATE     :  November 16, 1998
*
*
***************************************************************************/

#include "globals.h"

#if (LINUX_22)

ULONG remirror_active;
ULONG flush_active;
ULONG asynch_io_active[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };

ULONG exit_remirror;
ULONG exit_flush;
ULONG exit_asynch_io[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
ULONG asynch_signal[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
ULONG asynch_active[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
ULONG asynch_threads = 0;

struct semaphore mirror_sem = MUTEX_LOCKED;
struct semaphore flush_sem = MUTEX_LOCKED;
struct semaphore asynch_io_sem0 = MUTEX_LOCKED;
struct semaphore asynch_io_sem1 = MUTEX_LOCKED;
struct semaphore asynch_io_sem2 = MUTEX_LOCKED;
struct semaphore asynch_io_sem3 = MUTEX_LOCKED;
struct semaphore asynch_io_sem4 = MUTEX_LOCKED;
struct semaphore asynch_io_sem5 = MUTEX_LOCKED;
struct semaphore asynch_io_sem6 = MUTEX_LOCKED;
struct semaphore asynch_io_sem7 = MUTEX_LOCKED;

struct semaphore *sem_table[8]={
   &asynch_io_sem0, &asynch_io_sem1, &asynch_io_sem2, &asynch_io_sem3,
   &asynch_io_sem4, &asynch_io_sem5, &asynch_io_sem6, &asynch_io_sem7
};

struct task_struct *remirror_task = 0;
struct task_struct *flush_task = 0;
struct task_struct *asynch_io_task[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
struct timer_list flush_timer = { NULL, NULL, 0, 0, 0 };

extern ULONG nwvp_remirror_poll(void);
extern void nwvp_quiese_remirroring(void);
extern void process_asynch_io(ULONG disk);
extern ULONG asynch_io_pending(ULONG disk);

int count = 0;

void NWFSStartRemirror(void)
{
   if (remirror_task != current)
      SignalSemaphore(&mirror_sem);
}

void NWFSFlush(void)
{
   if (flush_task != current)
      SignalSemaphore(&flush_sem);
}

void RunAsynchIOQueue(ULONG disk)
{
    register int i = (disk % 8);

    if (!asynch_threads)
    {
       asynch_active[i] = TRUE;
       process_asynch_io(i);
       asynch_active[i] = 0;
       asynch_signal[i] = 0;
    }

    if (asynch_io_task[i] != current)
    {
       if (!asynch_signal[i])
       {
	  if (!asynch_active[i])
	  {
	     asynch_signal[i] = TRUE;
             SignalSemaphore(sem_table[i]);
	  }
       }
    }
}

int nwfs_mirror(void *notused)
{
    register ULONG ccode, i;

    current->session = 1;
    current->pgrp = 1;

    sprintf(current->comm, "nwfs-remirror");
    remirror_task = current;

    remirror_active++;
    while (!exit_remirror)
    {
       ccode = down_interruptible(&mirror_sem);
       if (ccode == -EINTR)
       {
	  NWFSPrint("Shutting Down REMIRRORING Process ... ");
	  exit_remirror = TRUE;
	  nwvp_quiese_remirroring();
	  while (nwvp_remirror_poll() == 0);
	  NWFSPrint("done\n");

	  NWFSPrint("Shutting Down LRU Timer ... ");
	  del_timer(&flush_timer);
	  NWFSPrint("done\n");

	  NWFSPrint("Shutting Down LRU Process ... ");
	  exit_flush = TRUE;
	  while (flush_active)
	  {
	     NWFSFlush();
#if (LINUX_SLEEP)
	     schedule();
#endif
	  }
	  NWFSPrint("done\n");

	  NWFSPrint("Shutting Down ASYNCH_IO Processes [ ");
          for (i=0; i < 8; i++)
          {
	     exit_asynch_io[i] = TRUE;
	     while (asynch_io_active[i])
	     {
	        RunAsynchIOQueue(i);
#if (LINUX_SLEEP)
	        schedule();
#endif
	     }
             NWFSPrint("%d ", (int)i);
          }
          NWFSPrint("]\n");
	  break;
       }

       if (exit_remirror)
       {
	   nwvp_quiese_remirroring();
	   while (nwvp_remirror_poll() == 0);
	   break;
       }

       while (nwvp_remirror_poll() == 0)
       {
	  if (exit_remirror)
	  {
	     nwvp_quiese_remirroring();
	     while (nwvp_remirror_poll() == 0);
	     break;
	  }
       }
#if (LINUX_SLEEP)
       schedule();
#endif
    }
    remirror_active--;
    return 0;

}

int nwfs_flush(void *notused)
{
    current->session = 1;
    current->pgrp = 1;

    sprintf(current->comm, "nwfs-flush");
    flush_task = current;

    flush_active++;
    while (!exit_flush)
    {
       down(&flush_sem);

       if (exit_flush)
	  break;

       if (DataLRU.Signal)
       {
	  if (!DataLRU.Active)
	     FlushEligibleLRU(&DataLRU);
	  DataLRU.Signal = 0;
       }
#if (LINUX_SLEEP)
       schedule();
#endif
    }
    flush_active--;
    return 0;

}

int nwfs_asynch_io_process(void *id)
{
    register int i = (int)id;

    current->session = 1;
    current->pgrp = 1;

    sprintf(current->comm, "nwfs-async%d", (int)i);
    asynch_io_task[i] = current;

    asynch_io_active[i]++;
    asynch_threads++;
    while (!exit_asynch_io[i])
    {
       if (!asynch_io_pending(i))
          down(sem_table[i]);

       if (exit_asynch_io[i])
	  break;

       if (asynch_signal[i])
       {
          if (!asynch_active[i])
          {
             asynch_active[i] = TRUE;
             process_asynch_io(i);
             asynch_active[i] = 0;
          }
          asynch_signal[i] = 0;
       }
#if (LINUX_SLEEP)
       schedule();
#endif
    }
    asynch_threads--;
    asynch_io_active[i]--;
    return 0;

}

void LRUTimer(void)
{
    flush_timer.data = 0;
    flush_timer.expires = jiffies + HZ; // second delay
    flush_timer.function = (void (*)(ULONG))LRUTimer;
    add_timer(&flush_timer);

    if (flush_task != current)
    {
       if (!DataLRU.Signal)
       {
	  if (!DataLRU.Active)
	  {
	     DataLRU.Signal = TRUE;
	     NWFSFlush();
          }
       }
    }
    return;
}

struct super_operations nwfs_sops =
{
    nwfs_read_inode,        // read inode
    nwfs_write_inode,	    // write inode
    nwfs_put_inode,	    // put inode
    nwfs_delete_inode,      // delete inode
    NULL,                   // notify change
    nwfs_put_super,	    // put superblock
    NULL,		    // write superblock
    nwfs_statfs,	    // stat filesystem
    NULL                    // remount_fs
};

struct file_system_type nwfs_fs_type =
{
    "nwfs",
    0,
    nwfs_read_super,
    NULL
};

int init_module(void)
{
    register int status, i;

    if ((status = register_filesystem(&nwfs_fs_type)) == 0)
       NWFSPrint("nwfs:  initialized successfully\n");

    InitNWFS();

    remirror_active = 0;
    flush_active = 0;

    exit_remirror = 0;
    exit_flush = 0;

    NWFSPrint("starting REMIRROR process ...\n");
    kernel_thread(nwfs_mirror, NULL, 0);

    NWFSPrint("starting LRU process ...\n");
    kernel_thread(nwfs_flush, NULL, 0);

    NWFSPrint("starting ASYNCH_IO processes [ ");
    for (i=0; i < 8; i++)
    {
       kernel_thread(nwfs_asynch_io_process, (void *)i, 0);
       NWFSPrint("%d ", (int)i);
    }
    NWFSPrint("]\n");

    NWFSPrint("starting LRU Timer ...\n");
    init_timer(&flush_timer);
    flush_timer.data = 0;
    flush_timer.expires = jiffies + HZ; // second delay
    flush_timer.function = (void (*)(ULONG))LRUTimer;
    add_timer(&flush_timer);

    NWFSVolumeScan();
    ReportVolumes();

    return status;
}

void cleanup_module(void)
{
    register int err, i;
    extern void displayMemoryList(void);

    err = unregister_filesystem(&nwfs_fs_type);
    if (err)
    {
       NWFSPrint("nwfs:  filesystem in use.  unload failed\n");
       return;
    }

    DismountAllVolumes();

    NWFSPrint("Shutting Down LRU Timer ... ");
    del_timer(&flush_timer);
    NWFSPrint("done\n");

    NWFSPrint("Shutting Down REMIRRORING Process ... ");
    exit_remirror = TRUE;
    while (remirror_active)
    {
       NWFSStartRemirror();
#if (LINUX_SLEEP)
       schedule();
#endif
    }
    NWFSPrint("done\n");

    NWFSPrint("Shutting Down LRU Process ... ");
    exit_flush = TRUE;
    while (flush_active)
    {
       NWFSFlush();
#if (LINUX_SLEEP)
       schedule();
#endif
    }
    NWFSPrint("done\n");

    NWFSPrint("Shutting Down ASYNCH_IO Processes [ ");
    for (i=0; i < 8; i++)
    {
       exit_asynch_io[i] = TRUE;
       while (asynch_io_active[i])
       {
	  RunAsynchIOQueue(i);
#if (LINUX_SLEEP)
	  schedule();
#endif
       }
       NWFSPrint("%d ", (int)i);
    }
    NWFSPrint("]\n");

    NWFSVolumeClose();
    RemoveDiskDevices();
    FreePartitionResources();

}

#endif


#if (LINUX_20)

ULONG remirror_active;
ULONG flush_active;
ULONG asynch_io_active[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };

ULONG exit_remirror;
ULONG exit_flush;
ULONG exit_asynch_io[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
ULONG asynch_signal[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
ULONG asynch_active[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
ULONG asynch_threads = 0;

struct semaphore mirror_sem = MUTEX_LOCKED;
struct semaphore flush_sem = MUTEX_LOCKED;
struct semaphore asynch_io_sem0 = MUTEX_LOCKED;
struct semaphore asynch_io_sem1 = MUTEX_LOCKED;
struct semaphore asynch_io_sem2 = MUTEX_LOCKED;
struct semaphore asynch_io_sem3 = MUTEX_LOCKED;
struct semaphore asynch_io_sem4 = MUTEX_LOCKED;
struct semaphore asynch_io_sem5 = MUTEX_LOCKED;
struct semaphore asynch_io_sem6 = MUTEX_LOCKED;
struct semaphore asynch_io_sem7 = MUTEX_LOCKED;

struct semaphore *sem_table[8]={
   &asynch_io_sem0, &asynch_io_sem1, &asynch_io_sem2, &asynch_io_sem3,
   &asynch_io_sem4, &asynch_io_sem5, &asynch_io_sem6, &asynch_io_sem7
};

struct task_struct *remirror_task = 0;
struct task_struct *flush_task = 0;
struct task_struct *asynch_io_task[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
struct timer_list flush_timer = { NULL, NULL, 0, 0, 0 };

extern ULONG nwvp_remirror_poll(void);
extern void nwvp_quiese_remirroring(void);
extern void process_asynch_io(ULONG disk);
extern ULONG asynch_io_pending(ULONG disk);

int count = 0;


void NWFSStartRemirror(void)
{
   if (remirror_task != current)
      SignalSemaphore(&mirror_sem);
}

void NWFSFlush(void)
{
   if (flush_task != current)
      SignalSemaphore(&flush_sem);
}

void RunAsynchIOQueue(ULONG disk)
{
    register int i = (disk % 8);

    if (!asynch_threads)
    {
       asynch_active[i] = TRUE;
       process_asynch_io(i);
       asynch_active[i] = 0;
       asynch_signal[i] = 0;
    }

    if (asynch_io_task[i] != current)
    {
       if (!asynch_signal[i])
       {
	  if (!asynch_active[i])
	  {
	     asynch_signal[i] = TRUE;
             SignalSemaphore(sem_table[i]);
	  }
       }
    }
}

int nwfs_mirror(void *notused)
{
    register ULONG ccode, i;

    current->session = 1;
    current->pgrp = 1;

    sprintf(current->comm, "nwfs-remirror");
    remirror_task = current;

    remirror_active++;
    while (!exit_remirror)
    {
       ccode = down_interruptible(&mirror_sem);
       if (ccode == -EINTR)
       {
	  NWFSPrint("Shutting Down REMIRRORING Process ... ");
	  exit_remirror = TRUE;
	  nwvp_quiese_remirroring();
	  while (nwvp_remirror_poll() == 0);
	  NWFSPrint("done\n");

	  NWFSPrint("Shutting Down LRU Timer ... ");
	  del_timer(&flush_timer);
	  NWFSPrint("done\n");

	  NWFSPrint("Shutting Down LRU Process ... ");
	  exit_flush = TRUE;
	  while (flush_active)
	  {
	     NWFSFlush();
#if (LINUX_SLEEP)
	     schedule();
#endif
	  }
	  NWFSPrint("done\n");

	  NWFSPrint("Shutting Down ASYNCH_IO Processes [ ");
          for (i=0; i < 8; i++)
          {
	     exit_asynch_io[i] = TRUE;
	     while (asynch_io_active[i])
	     {
	        RunAsynchIOQueue(i);
#if (LINUX_SLEEP)
	        schedule();
#endif
	     }
             NWFSPrint("%d ", (int)i);
          }
          NWFSPrint("]\n");
	  break;
       }

       if (exit_remirror)
       {
	   nwvp_quiese_remirroring();
	   while (nwvp_remirror_poll() == 0);
	   break;
       }

       while (nwvp_remirror_poll() == 0)
       {
	  if (exit_remirror)
	  {
	     nwvp_quiese_remirroring();
	     while (nwvp_remirror_poll() == 0);
	     break;
	  }
       }
#if (LINUX_SLEEP)
       schedule();
#endif
    }
    remirror_active--;
    return 0;

}

int nwfs_flush(void *notused)
{
    current->session = 1;
    current->pgrp = 1;

    sprintf(current->comm, "nwfs-flush");
    flush_task = current;

    flush_active++;
    while (!exit_flush)
    {
       down(&flush_sem);

       if (exit_flush)
	  break;

       if (DataLRU.Signal)
       {
	  if (!DataLRU.Active)
	     FlushEligibleLRU(&DataLRU);
          DataLRU.Signal = 0;
       }
#if (LINUX_SLEEP)
       schedule();
#endif
    }
    flush_active--;
    return 0;

}

int nwfs_asynch_io_process(void *id)
{
    register int i = (int)id;

    current->session = 1;
    current->pgrp = 1;

    sprintf(current->comm, "nwfs-async%d", (int)i);
    asynch_io_task[i] = current;

    asynch_io_active[i]++;
    asynch_threads++;
    while (!exit_asynch_io[i])
    {
       if (!asynch_io_pending(i))
          down(sem_table[i]);

       if (exit_asynch_io[i])
	  break;

       if (asynch_signal[i])
       {
          if (!asynch_active[i])
          {
             asynch_active[i] = TRUE;
             process_asynch_io(i);
             asynch_active[i] = 0;
          }
          asynch_signal[i] = 0;
       }
#if (LINUX_SLEEP)
       schedule();
#endif
    }
    asynch_threads--;
    asynch_io_active[i]--;
    return 0;

}

void LRUTimer(void)
{
    flush_timer.data = 0;
    flush_timer.expires = jiffies + HZ; // second delay
    flush_timer.function = (void (*)(ULONG))LRUTimer;
    add_timer(&flush_timer);

    if (flush_task != current)
    {
       if (!DataLRU.Signal)
       {
	  if (!DataLRU.Active)
	  {
	     DataLRU.Signal = TRUE;
	     NWFSFlush();
          }
       }
    }
    return;
}

struct file_system_type nwfs_type =
{
    nwfs_read_super, "nwfs", 0, NULL
};

struct super_operations nwfs_sops =
{
    nwfs_read_inode,      // read inode
    NULL,                   // notify change
    nwfs_write_inode,	    // write inode
    nwfs_put_inode,	    // put inode
    nwfs_put_super,	    // put superblock
    NULL,		    // write superblock
    nwfs_statfs,	    // stat filesystem
    NULL
};

int init_module(void)
{
    register int status, i;
    extern int nwfs_register_symbols(void);

    if ((status = register_filesystem(&nwfs_type)) == 0)
    {
       NWFSPrint("nwfs:  initialized successfully\n");
       nwfs_register_symbols();
    }

    InitNWFS();

    remirror_active = 0;
    flush_active = 0;

    exit_remirror = 0;
    exit_flush = 0;

    NWFSPrint("starting REMIRROR process ...\n");
    kernel_thread(nwfs_mirror, NULL, 0);

    NWFSPrint("starting LRU process ...\n");
    kernel_thread(nwfs_flush, NULL, 0);

    NWFSPrint("starting ASYNCH_IO processes [ ");
    for (i=0; i < 8; i++)
    {
       kernel_thread(nwfs_asynch_io_process, (void *)i, 0);
       NWFSPrint("%d ", (int)i);
    }
    NWFSPrint("]\n");

    NWFSPrint("starting LRU Timer ...\n");
    init_timer(&flush_timer);
    flush_timer.data = 0;
    flush_timer.expires = jiffies + HZ; // second delay
    flush_timer.function = (void (*)(ULONG))LRUTimer;
    add_timer(&flush_timer);

    NWFSVolumeScan();
    ReportVolumes();

    return status;
}

void cleanup_module(void)
{
    register int err, i;
    extern void displayMemoryList(void);

    err = unregister_filesystem(&nwfs_type);
    if (err)
    {
       NWFSPrint("nwfs:  filesystem in use.  unload failed\n");
       return;
    }

    DismountAllVolumes();

    NWFSPrint("Shutting Down LRU Timer ... ");
    del_timer(&flush_timer);
    NWFSPrint("done\n");

    NWFSPrint("Shutting Down REMIRRORING Process ... ");
    exit_remirror = TRUE;
    while (remirror_active)
    {
       NWFSStartRemirror();
#if (LINUX_SLEEP)
       schedule();
#endif
    }
    NWFSPrint("done\n");

    NWFSPrint("Shutting Down LRU Process ... ");
    exit_flush = TRUE;
    while (flush_active)
    {
       NWFSFlush();
#if (LINUX_SLEEP)
       schedule();
#endif
    }
    NWFSPrint("done\n");

    NWFSPrint("Shutting Down ASYNCH_IO Processes [ ");
    for (i=0; i < 8; i++)
    {
       exit_asynch_io[i] = TRUE;
       while (asynch_io_active[i])
       {
	  RunAsynchIOQueue(i);
#if (LINUX_SLEEP)
	  schedule();
#endif
       }
       NWFSPrint("%d ", (int)i);
    }
    NWFSPrint("]\n");

    NWFSVolumeClose();
    RemoveDiskDevices();
    FreePartitionResources();

}

#endif

